from typing import Type, Optional, List, Tuple, Dict, Union, Self, Generic, TypeVar

from NonameOrm.DB.DB import DB
from NonameOrm.Model.ModelExcutor import AsyncModelExecutor, ModelExecutor, BaseModelExecutor, Warp
from NonameOrm.Model.ModelProperty import BaseProperty, ForeignKey
from pydantic import BaseModel


class ModelInstance(dict):
    object: DataModel

    def __init__(self, *args, **kwargs): ...

    def __getattribute__(self, name): ...

    def __setattr__(self, key, value): ...


class _DataModel:
    @classmethod
    def instanceBuilder(cls: DataModel, *args, **kwargs) -> ModelInstance: ...


class _DataModelMeta(type):
    def __new__(cls, name, bases: tuple, attrs: dict, **kwargs): ...


class DataModel(_DataModel, ModelInstance, metaclass=_DataModelMeta):
    col: Tuple[BaseProperty]
    mapping: Dict[str, BaseProperty]
    tableName: str
    _db: Type[DB] = DB
    pkName: Optional[str] = None
    pkCol: Optional[BaseProperty] = None
    fk: List[ForeignKey]
    MODEL: Union[Type[DataModel], Type[BaseModel], Type[Self]]

    def __new__(cls, *args, **kwargs) -> ModelInstance: ...

    @classmethod
    def getAsyncExecutor(cls, work=None) -> AsyncModelExecutor['DataModel', Type[Self]]: ...

    @classmethod
    def getExecutor(cls, work=None) -> Union[AsyncModelExecutor['DataModel', Type[Self]], ModelExecutor['DataModel', Type[Self]]]: ...

T = TypeVar('T')

def get_executor(target:T, work=None) -> Union[AsyncModelExecutor, ModelExecutor, Warp[T]]:...



from typing import Callable, List, Any, Optional, Self, Union, TypeVar, Generic

from NonameOrm.DB.DB import DB
from NonameOrm.DB.Generator import SqlGenerator
from NonameOrm.Model.DataModel import InstanceList, ModelInstance, DataModel
from NonameOrm.Model.ModelProperty import BaseProperty, FilterListCell

T = TypeVar('T')

class Warp(Generic[T]):

    a = T
    def findAnyMatch(self, instance: ModelInstance, limit=None, offset=0) -> List[T]: ...

    def findAllBy(self, Filter: Union[bool, FilterListCell]) -> List[T]: ...

class BaseModelExecutor(Generic[T]):
    sql: SqlGenerator
    model: T
    db: DB
    work: Any

    def __init__(self, model: T, work=None): ...

    def instanceToFilter(self, instance: ModelInstance): ...

    def reset(self) -> Self: ...

    @property
    def generator(self) -> SqlGenerator: ...

    def findAnyMatch(self, instance: ModelInstance, limit=None, offset=0) -> List[T]: ...

    def find(self, *cols: List[BaseProperty]): ...

    def save(self, instance: ModelInstance): ...

    def delete(self, instance: Optional[ModelInstance] = None, filter: Optional[Union[bool, FilterListCell]] = None): ...

    def update(self, instance: ModelInstance): ...

    def findOneBy(self, Filter: Union[bool, FilterListCell]) -> T: ...

    def findAllBy(self, Filter: Union[bool, FilterListCell]) -> List[T]: ...

    def editSql(self, callback: Callable[[SqlGenerator], None]) -> Self: ...

    def execute(self): ...

    def process(self, res): ...



class ModelExecutor(BaseModelExecutor, Generic[T]):

    def __init__(self, model: T, work=None):
        super().__init__(model, work)


class AsyncModelExecutor(BaseModelExecutor, Generic[T]):

    def __init__(self, model: T, work=None):
        super().__init__(model, work)

    async def findAnyMatch(self, instance: ModelInstance, limit=None, offset=0) -> List[T]: ...

    async def findAllBy(self, Filter: Union[bool, FilterListCell] = None) -> List[T]: ...

    def find(self, *cols: List[BaseProperty]) -> AsyncModelExecutor: ...

    async def By(self, Fileter: FilterListCell) -> ModelInstance: ...

    async def save(self, instance: ModelInstance) -> T: ...

    async def findOneBy(self, Filter: Union[bool, FilterListCell]) -> T: ...

    async def update(self, instance: ModelInstance): ...

    async def execute(self) -> ModelInstance: ...

    async def findForeignKey(self, instance, deep: Optional[bool] = False): ...

    async def findListForeignKey(self, instances: InstanceList, deep: Optional[bool] = False): ...



from typing import List, Optional, TypedDict, Union, Awaitable, Self, Callable, Type, Any, TypeVar, Generic

from NonameOrm.DB.Generator import SqlGenerator
from NonameOrm.Model.DataModel import DataModel, ModelInstance
from NonameOrm.Model.ModelProperty import FilterListCell

from NonameOrm.Ext.Dict import DictPlus


class _Page(TypedDict):
    page: int
    pageSize: int
    content: List[Union[DataModel, ModelInstance]]
    total: int
    totalPage: int

T = TypeVar('T')

class Page(dict, Generic[T]):
    """
    Page实例类
    本类继承自dict，拥有字典的一切行为

    额外支持通过 . 获取字典内容
    若获取的内容与字典类成员方法,参数相冲突，则优先选取字典内元素
    """
    page: int
    pageSize: int
    content: List[Union[T, ModelInstance]]
    total: int
    totalPage: int

    def __getattr__(self, item: str): ...

    def __getitem__(self, item: str): ...

    def __setattr__(self, key: str, value: Any): ...

    def __init__(self, page: int, pageSize: int, content=None, total: int = 0):
        pass





class _PageAble(Generic[T]):
    """
    分页控制器

    本类支持链式调用
    """

    def __init__(self, target: T, page: Optional[int] = 1, pageSize: Optional[int] = 10, findForeign: Optional[bool] = False):
        """

        :param target:
        :param page:
        :param findForeign:
        :param pageSize:
        """
        pass

    def filter(self, args: Optional[FilterListCell] = None) -> Union["_PageAble"[T], Self[T]]:
        """
        设置过滤条件

        :param args: FilterListCell
        :return: PageAble
        """
        pass

    def findForeign(self) -> 'PageAble':
        pass

    def orderBy(self, *order: List[str]) -> 'PageAble':
        pass

    def setPage(self, page: int, pageSize: Optional[int] = 0) -> Self:
        """
        设置分页

        :param page: 当前页数
        :param pageSize: 页面大小 默认可不传
        :return: PageAble
        """
        pass

    def editSql(self, callable: Callable[[SqlGenerator], None]) -> Self: ...

    @property
    def sql(self) -> SqlGenerator: ...

    def execute(self) -> Union[Page[T], Awaitable[Page[T]]]:
        """
        链式调用尽头
        正式进行数据获取
        """
        pass

def PageAble(target: T, page: int = 1, pageSize: int = 10, findForeign: bool = False) -> _PageAble[T]: ...



from .BaseOrmError import *


class DBInstanceError(BaseOrmError):
    msg: str = "DB链接对象实例已存在，不可重新创建"


class DBInstanceCreateError(BaseOrmError):
    msh: str = "请使用DB.create()创建数据库实例"


class WriteOperateNotInAffairs(BaseOrmError):
    msg: str = "写操作需要在事务中执行"


class UpdateWithOutPrimaryKeyError(BaseOrmError):
    msg = "更新数据库实例需要包含主键字段"


class DeleteWithOutPrimaryKeyError(BaseOrmError):
    msg = "删除数据库实例需要包含主键字段"


class InOperatorError(BaseOrmError):
    msg = "In操作符传入了非法类型"



from asyncio import Task, AbstractEventLoop
from typing import Optional, Any, Awaitable, Dict, Union

from aiomysql import Connection as AsyncMysqlConnection

from aiosqlite.core import Connection as AsyncSqliteConnection

from NonameOrm.Model.ModelExcutor import AsyncModelExecutor


class BaseConnector:
    """
        数据库连接器基类，无任何功能，需要继承实现
    """
    isAsync: bool

    def getCon(self): ...

    def execute(self, sql: str, con: Any = None, args: tuple = ()): ...

    def process(self, *args, **kwargs):
        pass

    def releaseCon(self, con: Any): ...

    def asyncProcess(self, *args, **kwargs):
        pass


class AioSqliteConnector(BaseConnector):
    """
        异步AioSqlite连接器

        基于 aiosqlite 封装
    """

    conMap: Dict[Task:AsyncSqliteConnection]

    def __init__(self, loop: AbstractEventLoop, path: str, showLog: bool = True): ...

    async def getCon(self) -> AsyncSqliteConnection: ...

    async def execute(self, sql: str, con: Optional[AsyncSqliteConnection] = None, args: tuple = ()): ...

    def releaseCon(self, con: AsyncMysqlConnection) -> Awaitable: ...

    async def asyncProcess(self,
                           executor: AsyncModelExecutor,
                           ): ...


class Sqlite3Connector(BaseConnector):
    def __init__(self, path: str, showLog: bool = True): ...


class AioMysqlConnector(BaseConnector):
    """
        异步mysql 连接器

        基于 aiomysql 封装
    """

    def __init__(self, *args, **kwargs): ...

    @property
    def config(self) -> dict: ...

    async def getCon(self) -> AsyncMysqlConnection: ...

    def releaseCon(self, con: AsyncMysqlConnection) -> Awaitable: ...

    async def execute(self, sql: str, con: AsyncMysqlConnection = None, args: tuple = ()): ...

    async def asyncProcess(self, *args, **kwargs): ...



from enum import Enum
from typing import Any, TypedDict, Optional, Type, Iterable, List, TypeVar

from NonameOrm.DB.Generator import CustomColAnnounce
from NonameOrm.Model.ModelProperty import FilterListCell

from NonameOrm.Model import DataModel
from NonameOrm.Model.DataModel import MiddleDataModel


class NullDefault:
    pass


class AutoIncrement:
    pass


_null: NullDefault = NullDefault()

auto_increment: Type[AutoIncrement] = AutoIncrement

current_timestamp = CustomColAnnounce("current_timestamp")
current_date = CustomColAnnounce("current_date")


class InsertCell(TypedDict):
    name: str
    value: Any


class NullDefault: ...


_null: NullDefault


class BaseProperty:
    name: str
    _Type: type
    isPk: bool
    Null: bool
    _default: Any
    targetType: Any
    model: DataModel

    def __init__(self,
                 name: str = None,
                 pk: bool = False,
                 default: Any = _null,
                 Null: bool = True,
                 typeArgs: tuple = (),
                 targetType: Enum = None,
                 *args, **kwargs): ...

    def _init(*args, **kwargs): ...

    def __set_name__(self, owner, name: str) -> None: ...

    def sizeChecker(self, value: Any) -> bool: ...

    def verifier(self, value: Any) -> bool: ...

    def insertCell(self, value) -> InsertCell: ...

    def updateCell(self, value) -> InsertCell: ...

    def toDBValue(self, value) -> Any: ...

    def desc(self) -> str: ...

    def asc(self) -> str: ...

    def toObjValue(self, value): ...

    def __eq__(self, other) -> FilterListCell: ...

    def __ne__(self, other) -> FilterListCell: ...

    def __lt__(self, other) -> FilterListCell: ...

    def __gt__(self, other) -> FilterListCell: ...

    def __le__(self, other) -> FilterListCell: ...

    def __ge__(self, other) -> FilterListCell: ...

    def within(self, other: Iterable) -> FilterListCell: ...

    def is_null(self) -> FilterListCell: ...

    # def setFilter(self, other, relationship) -> FilterListCell: ...

    @property
    def hasDefault(self) -> bool: ...

    @property
    def Default(self) -> Any: ...

    @property
    def dbName(self) -> str: ...

    @property
    def objName(self) -> str: ...


class IntProperty(BaseProperty, int): ...


class StrProperty(BaseProperty, str):
    def like(self, value: str) -> FilterListCell: ...

    def startsWith(self, value: str) -> FilterListCell: ...

    def endsWith(self, value: str) -> FilterListCell: ...

    def has(self, value: str) -> FilterListCell: ...


class FloatProperty(BaseProperty, float): ...


class BoolProperty(BaseProperty, bool): ...


class JsonProperty(BaseProperty, dict, list): ...


class TimestampProperty(BaseProperty, str):
    @property
    def date(self) -> FilterListCell: ...

    ...


class timestampSupportType(Enum):
    varchar = 'varchar(25)'
    timestamp = 'timestamp'
    datetime = 'datetime'


class floatSupportType(Enum):
    float = 'float'
    decimal = 'decimal'


class boolSupportType(Enum):
    tinyInt = 'tinyint'
    bit = 'bit'
    varchar = 'varchar(5)'


class jsonSupportType(Enum):
    varchar = 'varchar(255)'
    text = 'text'
    longtext = 'longtext'
    tinytext = 'tinytext'
    json = 'json'


class strSupportType(Enum):
    varchar = 'varchar'
    text = 'text'
    longText = 'longtext'
    tinyText = 'tinytext'


class ForeignType(Enum):
    ONE_TO_ONE = 1
    ONE_TO_MANY = 2
    MANY_TO_MANY = 3


T = TypeVar('T')


class ForeignKey(dict, List[T]):
    from NonameOrm import Model
    target: Model.DataModel.DataModel
    bindCol: BaseProperty
    Type: ForeignType
    owner: Model.DataModel.DataModel
    targetBindCol: Optional[BaseProperty]
    middleModel: Optional[MiddleDataModel]
    name: str

    def __init__(self, target: T, Type: ForeignType = ForeignType.ONE_TO_ONE, bindCol: Optional[BaseProperty] = None, targetBindCol: Optional[BaseProperty] = None): ...

    def __set_name__(self, owner, name): ...

    def __getattribute__(self, item: str): ...

    @property
    def directTarget(self) -> Model.DataModel.DataModel: ...



from typing import Optional, Type, List, Union, TypeVar

from NonameOrm.DB.DB cimport DB
from NonameOrm.Model.ModelExcutor cimport AsyncModelExecutor
from NonameOrm.Model.ModelExcutor import ModelExecutor
from NonameOrm.Model.ModelProperty import ForeignKey, ForeignType
from pydantic import create_model

from NonameOrm.Ext.Decorators import pydantic_support
from .ModelProperty cimport *

cdef str get_lower_case_name(str text):
    text = text.replace('Model', '')
    cdef list lst = []
    cdef int index
    cdef str char
    for index, char in enumerate(text):
        if char.isupper() and index != 0:
            lst.append("_")
        lst.append(char)

    return "".join(lst).lower()

cdef class _DataModel:
    @classmethod
    def instanceBuilder(cls: DataModel, *args, **kwargs) -> ModelInstance:
        return cls.modelInstance(*args, **kwargs)


class _DataModelMeta(type):
    """
        数据模型元类
        在生成数据模型子类时，将会在这对模型进行预处理操作
    """
    ModelList: List['DataModel'] = []  # 模型列表，存放所定义的所有模型

    def __new__(cls, str name, bases: tuple, attrs: dict, **kwargs):
        # 根据模型类定义，得到表的定义内容，并进行归总
        attrs['col'], attrs['mapping'], attrs['fk'] = _DataModelMeta.getPropertyObj(cls, attrs)

        # 判断当前模型是否定义了表名，若没有则根据模型类名生成表名
        # 生成规则如下，去除Model结尾，并将名称转为蛇型命名
        # 例子：UserWalletModel -> user_wallet
        if not attrs.get('tableName'):
            attrs['tableName'] = get_lower_case_name(name)

        # 根据所得的定义，生成模型类
        Class = type.__new__(cls, name, bases, attrs)

        # 将生成的模型类添加进模型列表
        _DataModelMeta.ModelList.append(Class)

        # 往模型里注入相关的内容
        # 注入模型实例类到模型类的  modelInstance  成员中
        setattr(Class, 'modelInstance', _DataModelMeta.buildModelInstance(Class, attrs['col'], name))
        # 注入Pydantic模型到模型类的  MODEL  成员中
        setattr(Class, 'MODEL', _DataModelMeta.buildPydanticModel(Class, attrs['col'], name, attrs['fk']))
        return Class

    @staticmethod
    def buildModelInstance(cls, list cols: List[BaseProperty], str name) -> Type[ModelInstance]:
        """
            在这里会生成模型的 模型实例类
        """
        cdef dict temp = dict()

        return type(name + 'Instance', (ModelInstance,), {'object': cls, '_temp': {}})

    @staticmethod
    def buildPydanticModel(cls, list cols: List[BaseProperty], str name, list fk: List[ForeignKey]):
        """
            以数据模型为基础
            生成Pydantic模型
        """
        cdef:
            BaseProperty prop
            dict types = {}
            int index
            list temp
        # 处理字段类型声明
        for index in range(len(cols)):
            prop = cols[index]
            temp = [Optional[prop._Type] if prop.Null else prop._Type]
            if prop.hasDefault:
                temp.append(prop.Default)
            else:
                temp.append(None)
            types[prop.name] = tuple(temp)

        # 处理外键类型声明
        for foregin in fk:
            if foregin['Type'] == ForeignType.ONE_TO_ONE:
                if foregin.target == cls:
                    types[foregin['name']] = (Optional[dict], None)
                else:
                    types[foregin['name']] = (Optional[foregin['target'].MODEL], None)
            else:
                if foregin.directTarget == cls:
                    types[foregin['name']] = (List[dict], [])
                else:
                    types[foregin['name']] = (List[foregin.directTarget.MODEL], [])
        Class = create_model(name, **types)
        # 往类里注入魔法方法，使其支持字典的部分特性
        pydantic_support(Class)
        return Class

    @staticmethod
    def getPropertyObj(type cls, attrs: dict):
        """
            根据表定义，将相关的property 和 外键
            进行归类并返回
            同时返回他们 名字 与 其本身的映射字典
        """
        cdef list property_list = []  # 字段列表
        cdef list fk = []  # 外键列表
        cdef dict mapping = {}  # 映射字典
        cdef str key

        for key, item in attrs.items():
            if isinstance(item, BaseProperty):
                property_list.append(item)
                mapping[key] = item
            elif isinstance(item, ForeignKey):
                fk.append(item)
                mapping[key] = item
        return property_list, mapping, fk


cdef class ModelInstance(dict):
    """
    模型实例基类
    其子类会在DataModel被创建之时自动创建并继承此类
    本类及其子类需要通过实例化模型类时实例化

    本类继承自字典《dict》实例拥有字典的所有能力
    遇到
    """

    def __init__(self, *args, **kwargs):
        cdef:
            str k
            cdef dict data = kwargs



        if len(args):
            if isinstance(args[0], zip):
                super().__init__(args[0])
            elif isinstance(args[0], dict):
                data = <dict> args[0]
                if kwargs.get('check', True):
                    super().__init__()
            else:
                super().__init__()

        from NonameOrm.Model.ModelProperty import ForeignType
        for k, v in self.object.mapping.items():
            if k in data:
                if isinstance(v, ForeignKey) and data[k] is not None:
                    if v.Type == ForeignType.ONE_TO_ONE:
                        self[k] = v.target(data[k], check=kwargs.get('check', True))
                    elif v.Type == ForeignType.ONE_TO_MANY:
                        self[k] = [v.target(i, check=kwargs.get('check', True)) for i in data[k]]
                    elif v.Type == ForeignType.MANY_TO_MANY:
                        self[k] = [v.directTarget(i, check=kwargs.get('check', True)) for i in data[k]]
                elif not kwargs.get('check', True):
                    self[k] = data[k]
                elif isinstance(v, BaseProperty):
                    self[k] = v.toObjValue(data[k])
            elif not kwargs.get('check', True):
                continue
            elif k in self:
                self[k] = v.toObjValue(self[k])
            elif isinstance(v, ForeignKey):
                continue
            else:
                self[v.name] = v.toObjValue(v.Default)

    def __getattribute__(self, str name):
        try:
            return self[name]
        except Exception:
            return object.__getattribute__(self, name)

    def __setattr__(self, key, value):
        if key in self.object.mapping:
            self[key] = value
            return
        raise KeyError()

cdef class InstanceList(list):
    """
    实例列表，内部只有一种对象，就是模型实例的实例
    """

    cdef int len(self):
        return len(self)


class DataModel(_DataModel, metaclass=_DataModelMeta):
    """
    本类为无实例类，无法实例化
    在调用本类的实例化方法时，将会创建本类相对应的模型实例类的实例即

    isinstance(DataModel(), DataModel) -> False
    isinstance(DataModel(), ModelInstance) -> True

    模型实例类 与 模型类 与 模型实例之间的关系需要注意
    """

    _db: Type[DB] = DB
    pkName: Optional[str] = None
    pkCol: Optional[BaseProperty] = None

    def __new__(cls, *args, **kwargs) -> ModelInstance:
        return cls.instanceBuilder(*args, **kwargs)

    @classmethod
    def getAsyncExecutor(cls, work=None) -> AsyncModelExecutor:
        return AsyncModelExecutor(cls, work)

    @classmethod
    def getExecutor(cls, work=None):
        if DB.getInstance().connector.isAsync:
            return cls.getAsyncExecutor(work)
        return ModelExecutor(cls, work)


class MiddleDataModel(DataModel):

    @classmethod
    def instanceBuilder(cls, *args, **kwargs) -> ModelInstance:
        pass

    @classmethod
    def getOtherModelBy(cls, model: DataModel) -> DataModel:
        return cls.getOtherFkBy(model).target

    @classmethod
    def getOtherFkBy(cls, model) -> ForeignKey:
        return cls.fk[0] if cls.fk[0].target is not model else cls.fk[1]


def get_executor(target, work=None):
    return target.getExecutor(work)


import logging
import time
from time import sleep

import aiosqlite
import faker

from NonameOrm.DB.Connector import AioMysqlConnector, AioSqliteConnector, Sqlite3Connector
from NonameOrm.DB.DB import DB
from NonameOrm.Ext.DataGenerator import data_task, GeneratorRunner
from NonameOrm.Ext.Decorators import use_database
from NonameOrm.Model.DataModel import DataModel
from NonameOrm.Model.ModelProperty import *
import asyncio
from faker import Faker

from NonameOrm.DB.Generator import SqlGenerator
from test.Models import *

loop = asyncio.new_event_loop()
asyncio.set_event_loop(loop)

from NonameOrm.Ext.PageAble import PageAble

fake = Faker(loacale='zh_CN')

logging.basicConfig(level=logging.DEBUG, format='[ %(levelname)s - %(pathname)s - %(funcName)s] %(message)s')
logger = logging.getLogger(__name__)

if __name__ == '__main__':
    # DB.create(connector=Sqlite3Connector('test.db')).GenerateTable()
    # # print(Task.watch_sku_list.Type)
    # # print(Task.watch_sku_list.Type.__class__)
    # Task.getExecutor().save(Task(name='test',product_id='123'))
    # for i in range(10000):
    #     Task.getExecutor().findOneBy(Task.id == 1)
    #
    # sleep(100)
    from NonameOrm.DB.Generator import Q


    # genertor = SqlGenerator(Task)


import functools
import inspect
from typing import Awaitable

from NonameOrm.DB.DB import DB
from pydantic import BaseModel


def pydantic_support(obj: BaseModel) -> BaseModel:
    """
        装饰器，也可当函数使用
        装饰后的pydantic模型可以被 NonameOrm 所使用
    """
    setattr(obj, '__getitem__', lambda self, item: getattr(self, item))  # 使模型可以通过['xxx']访问成员
    setattr(obj, '__setitem__', lambda self, item, value: setattr(self, item, value))  # 使模型可以通过 x['xx'] = xxx 来变更成员
    setattr(obj, 'get', lambda self, item, default: getattr(self, item))  # 使模型可以使用get('xxx',xxx)获取成员
    setattr(obj, '__contains__', lambda self, item: hasattr(self, item))  # 使模型可以使用 in 操作附判断是否含有成员
    return obj


def use_database(fun):
    if inspect.iscoroutinefunction(fun):
        @functools.wraps(fun)
        async def wrap(*args, **kwargs):
            con = await DB.getInstance().connector.getCon()
            try:
                return await fun(*args, **kwargs)
            except Exception as e:
                await con.rollback()
                raise e
    else:
        @functools.wraps(fun)
        def wrap(*args, **kwargs):
            con = DB.getInstance().connector.getCon()
            try:
                return fun(*args, **kwargs)
            except Exception as e:
                con.rollback()
                raise e
    return wrap
